@@ -1,7 +1,7 @@
 use crate::{
     account::{ReadableAccount, WritableAccount},
     account_utils::State as AccountUtilsState,
-    ic_msg,
+    feature_set, ic_msg,
     keyed_account::KeyedAccount,
     nonce_account::create_account,
     process_instruction::InvokeContext,
@@ -10,7 +10,7 @@ use solana_program::{
     instruction::{checked_add, InstructionError},
     nonce::{self, state::Versions, State},
     pubkey::Pubkey,
-    system_instruction::NonceError,
+    system_instruction::{nonce_to_instruction_error, NonceError},
     sysvar::{recent_blockhashes::RecentBlockhashes, rent::Rent},
 };
 use std::collections::HashSet;
@@ -53,12 +53,17 @@ impl<'a> NonceKeyedAccount for KeyedAccount<'a> {
         signers: &HashSet<Pubkey>,
         invoke_context: &dyn InvokeContext,
     ) -> Result<(), InstructionError> {
+        let merge_nonce_error_into_system_error = invoke_context
+            .is_feature_active(&feature_set::merge_nonce_error_into_system_error::id());
         if recent_blockhashes.is_empty() {
             ic_msg!(
                 invoke_context,
                 "Advance nonce account: recent blockhash list is empty",
             );
-            return Err(NonceError::NoRecentBlockhashes.into());
+            return Err(nonce_to_instruction_error(
+                NonceError::NoRecentBlockhashes,
+                merge_nonce_error_into_system_error,
+            ));
         }
 
         let state = AccountUtilsState::<Versions>::state(self)?.convert_to_current();
@@ -78,7 +83,10 @@ impl<'a> NonceKeyedAccount for KeyedAccount<'a> {
                         invoke_context,
                         "Advance nonce account: nonce can only advance once per slot"
                     );
-                    return Err(NonceError::NotExpired.into());
+                    return Err(nonce_to_instruction_error(
+                        NonceError::NotExpired,
+                        merge_nonce_error_into_system_error,
+                    ));
                 }
 
                 let new_data = nonce::state::Data {
@@ -94,7 +102,10 @@ impl<'a> NonceKeyedAccount for KeyedAccount<'a> {
                     "Advance nonce account: Account {} state is invalid",
                     self.unsigned_key()
                 );
-                Err(NonceError::BadAccountState.into())
+                Err(nonce_to_instruction_error(
+                    NonceError::BadAccountState,
+                    merge_nonce_error_into_system_error,
+                ))
             }
         }
     }
@@ -108,6 +119,8 @@ impl<'a> NonceKeyedAccount for KeyedAccount<'a> {
         signers: &HashSet<Pubkey>,
         invoke_context: &dyn InvokeContext,
     ) -> Result<(), InstructionError> {
+        let merge_nonce_error_into_system_error = invoke_context
+            .is_feature_active(&feature_set::merge_nonce_error_into_system_error::id());
         let signer = match AccountUtilsState::<Versions>::state(self)?.convert_to_current() {
             State::Uninitialized => {
                 if lamports > self.lamports()? {
@@ -128,7 +141,10 @@ impl<'a> NonceKeyedAccount for KeyedAccount<'a> {
                             invoke_context,
                             "Withdraw nonce account: nonce can only advance once per slot"
                         );
-                        return Err(NonceError::NotExpired.into());
+                        return Err(nonce_to_instruction_error(
+                            NonceError::NotExpired,
+                            merge_nonce_error_into_system_error,
+                        ));
                     }
                     self.set_state(&Versions::new_current(State::Uninitialized))?;
                 } else {
@@ -180,12 +196,17 @@ impl<'a> NonceKeyedAccount for KeyedAccount<'a> {
         rent: &Rent,
         invoke_context: &dyn InvokeContext,
     ) -> Result<(), InstructionError> {
+        let merge_nonce_error_into_system_error = invoke_context
+            .is_feature_active(&feature_set::merge_nonce_error_into_system_error::id());
         if recent_blockhashes.is_empty() {
             ic_msg!(
                 invoke_context,
                 "Initialize nonce account: recent blockhash list is empty",
             );
-            return Err(NonceError::NoRecentBlockhashes.into());
+            return Err(nonce_to_instruction_error(
+                NonceError::NoRecentBlockhashes,
+                merge_nonce_error_into_system_error,
+            ));
         }
 
         match AccountUtilsState::<Versions>::state(self)?.convert_to_current() {
@@ -213,7 +234,10 @@ impl<'a> NonceKeyedAccount for KeyedAccount<'a> {
                     "Initialize nonce account: Account {} state is invalid",
                     self.unsigned_key()
                 );
-                Err(NonceError::BadAccountState.into())
+                Err(nonce_to_instruction_error(
+                    NonceError::BadAccountState,
+                    merge_nonce_error_into_system_error,
+                ))
             }
         }
     }
@@ -224,6 +248,8 @@ impl<'a> NonceKeyedAccount for KeyedAccount<'a> {
         signers: &HashSet<Pubkey>,
         invoke_context: &dyn InvokeContext,
     ) -> Result<(), InstructionError> {
+        let merge_nonce_error_into_system_error = invoke_context
+            .is_feature_active(&feature_set::merge_nonce_error_into_system_error::id());
         match AccountUtilsState::<Versions>::state(self)?.convert_to_current() {
             State::Initialized(data) => {
                 if !signers.contains(&data.authority) {
@@ -246,7 +272,10 @@ impl<'a> NonceKeyedAccount for KeyedAccount<'a> {
                     "Authorize nonce account: Account {} state is invalid",
                     self.unsigned_key()
                 );
-                Err(NonceError::BadAccountState.into())
+                Err(nonce_to_instruction_error(
+                    NonceError::BadAccountState,
+                    merge_nonce_error_into_system_error,
+                ))
             }
         }
     }
@@ -273,7 +302,7 @@ mod test {
         nonce::{self, State},
         nonce_account::verify_nonce_account,
         process_instruction::MockInvokeContext,
-        system_instruction::NonceError,
+        system_instruction::SystemError,
         sysvar::recent_blockhashes::create_test_recent_blockhashes,
     };
     use solana_program::hash::Hash;
@@ -456,7 +485,7 @@ mod test {
                 &signers,
                 &MockInvokeContext::new(vec![]),
             );
-            assert_eq!(result, Err(NonceError::NoRecentBlockhashes.into()));
+            assert_eq!(result, Err(SystemError::NonceNoRecentBlockhashes.into()));
         })
     }
 
@@ -485,7 +514,7 @@ mod test {
                 &signers,
                 &MockInvokeContext::new(vec![]),
             );
-            assert_eq!(result, Err(NonceError::NotExpired.into()));
+            assert_eq!(result, Err(SystemError::NonceBlockhashNotExpired.into()));
         })
     }
 
@@ -505,7 +534,7 @@ mod test {
                 &signers,
                 &MockInvokeContext::new(vec![]),
             );
-            assert_eq!(result, Err(NonceError::BadAccountState.into()));
+            assert_eq!(result, Err(InstructionError::InvalidAccountData));
         })
     }
 
@@ -858,7 +887,7 @@ mod test {
                     &signers,
                     &MockInvokeContext::new(vec![]),
                 );
-                assert_eq!(result, Err(NonceError::NotExpired.into()));
+                assert_eq!(result, Err(SystemError::NonceBlockhashNotExpired.into()));
             })
         })
     }
@@ -1024,7 +1053,7 @@ mod test {
                 &rent,
                 &MockInvokeContext::new(vec![]),
             );
-            assert_eq!(result, Err(NonceError::NoRecentBlockhashes.into()));
+            assert_eq!(result, Err(SystemError::NonceNoRecentBlockhashes.into()));
         })
     }
 
@@ -1053,7 +1082,7 @@ mod test {
                 &rent,
                 &MockInvokeContext::new(vec![]),
             );
-            assert_eq!(result, Err(NonceError::BadAccountState.into()));
+            assert_eq!(result, Err(InstructionError::InvalidAccountData));
         })
     }
 
@@ -1131,7 +1160,7 @@ mod test {
                 &signers,
                 &MockInvokeContext::new(vec![]),
             );
-            assert_eq!(result, Err(NonceError::BadAccountState.into()));
+            assert_eq!(result, Err(InstructionError::InvalidAccountData));
         })
     }
 
